"""Montreal Forced Aligner is a package for aligning speech corpora through the use of acoustic models and
            dictionaries using Kaldi functionality."""
import montreal_forced_aligner.aligner as aligner  # noqa
import montreal_forced_aligner.command_line as command_line  # noqa
import montreal_forced_aligner.config as config  # noqa
import montreal_forced_aligner.corpus as corpus  # noqa
import montreal_forced_aligner.dictionary as dictionary  # noqa
import montreal_forced_aligner.exceptions as exceptions  # noqa
import montreal_forced_aligner.g2p as g2p  # noqa
import montreal_forced_aligner.helper as helper  # noqa
import montreal_forced_aligner.models as models  # noqa
import montreal_forced_aligner.multiprocessing as multiprocessing  # noqa
import montreal_forced_aligner.textgrid as textgrid  # noqa
import montreal_forced_aligner.utils as utils  # noqa

__all__ = [
    "aligner",
    "command_line",
    "config",
    "corpus",
    "dictionary",
    "exceptions",
    "g2p",
    "helper",
    "models",
    "multiprocessing",
    "textgrid",
    "utils",
]



"""Class definitions for configuring sound file segmentation"""
from __future__ import annotations

import os

import yaml

from .base_config import BaseConfig, ConfigError
from .feature_config import FeatureConfig

__all__ = ["SegmentationConfig", "segmentation_yaml_to_config", "load_basic_segmentation"]


class SegmentationConfig(BaseConfig):
    """
    Class for storing segmentation configuration
    """

    def __init__(self, feature_config):
        self.use_mp = True
        self.energy_threshold = 5.5
        self.energy_mean_scale = 0.5
        self.max_segment_length = 30
        self.min_pause_duration = 0.05
        self.snap_boundary_threshold = 0.15
        self.feature_config = feature_config
        self.feature_config.use_energy = True
        self.overwrite = True

    def update(self, data: dict) -> None:
        """Update configuration parameters"""
        for k, v in data.items():
            if k == "use_mp":
                self.feature_config.use_mp = v
            if not hasattr(self, k):
                raise ConfigError("No field found for key {}".format(k))
            setattr(self, k, v)

    @property
    def segmentation_options(self):
        """Options for segmentation"""
        return {
            "max_segment_length": self.max_segment_length,
            "min_pause_duration": self.min_pause_duration,
            "snap_boundary_threshold": self.snap_boundary_threshold,
            "frame_shift": round(self.feature_config.frame_shift / 1000, 2),
        }


def segmentation_yaml_to_config(path: str) -> SegmentationConfig:
    """
    Helper function to load segmentation configurations

    Parameters
    ----------
    path: str
        Path to yaml file

    Returns
    -------
    :class:`~montreal_forced_aligner.config.segmentation_config.SegmentationConfig`
        Segmentation configuration
    """
    with open(path, "r", encoding="utf8") as f:
        data = yaml.load(f, Loader=yaml.SafeLoader)
        global_params = {}
        feature_config = FeatureConfig()
        for k, v in data.items():
            if k == "features":
                feature_config.update(v)
            else:
                global_params[k] = v
        segmentation_config = SegmentationConfig(feature_config)
        segmentation_config.update(global_params)
        return segmentation_config


def load_basic_segmentation() -> SegmentationConfig:
    """
    Helper function to load the default parameters

    Returns
    -------
    :class:`~montreal_forced_aligner.config.segmentation_config.SegmentationConfig`
        Default segmentation configuration
    """
    base_dir = os.path.dirname(os.path.abspath(__file__))
    segmentation_config = segmentation_yaml_to_config(
        os.path.join(base_dir, "basic_segmentation.yaml")
    )
    return segmentation_config



"""Class definitions for Speaker classification in MFA"""
from __future__ import annotations

import logging
import os
import shutil
from typing import TYPE_CHECKING, Optional

import numpy as np

from .config import TEMP_DIR
from .corpus.classes import Speaker
from .exceptions import KaldiProcessingError
from .helper import load_scp
from .multiprocessing import extract_ivectors
from .utils import log_kaldi_errors

if TYPE_CHECKING:
    from .config import SpeakerClassificationConfig
    from .corpus import TranscribeCorpus
    from .models import IvectorExtractor, MetaDict

__all__ = ["SpeakerClassifier"]


class SpeakerClassifier:
    """
    Class for performing speaker classification

    Parameters
    ----------
    corpus : :class:`~montreal_forced_aligner.corpus.base.Corpus`
        Corpus object for the dataset
    ivector_extractor : :class:`~montreal_forced_aligner.models.IvectorExtractor`
        Configuration for alignment
    classification_config : :class:`~montreal_forced_aligner.config.speaker_classification_config.SpeakerClassificationConfig`
        Configuration for alignment
    compute_segments: bool, optional
        Flag for whether segments should be created
    num_speakers: int, optional
        Number of speakers in the corpus, if known
    cluster: bool, optional
        Flag for whether speakers should be clustered instead of classified
    temp_directory : str, optional
        Specifies the temporary directory root to save files need for Kaldi.
        If not specified, it will be set to ``~/Documents/MFA``
    call_back : callable, optional
        Specifies a call back function for diarization
    debug : bool
        Flag for running in debug mode, defaults to false
    verbose : bool
        Flag for running in verbose mode, defaults to false
    """

    def __init__(
        self,
        corpus: TranscribeCorpus,
        ivector_extractor: IvectorExtractor,
        classification_config: SpeakerClassificationConfig,
        compute_segments: Optional[bool] = False,
        num_speakers: Optional[int] = None,
        cluster: Optional[bool] = False,
        temp_directory: Optional[str] = None,
        debug: Optional[bool] = False,
        verbose: Optional[bool] = False,
        logger: Optional[logging.Logger] = None,
    ):
        self.corpus = corpus
        self.ivector_extractor = ivector_extractor
        self.feature_config = self.ivector_extractor.feature_config
        self.classification_config = classification_config

        if not temp_directory:
            temp_directory = TEMP_DIR
        self.temp_directory = temp_directory
        os.makedirs(self.temp_directory, exist_ok=True)
        self.debug = debug
        self.compute_segments = compute_segments
        self.verbose = verbose
        if logger is None:
            self.log_file = os.path.join(self.temp_directory, "speaker_classifier.log")
            self.logger = logging.getLogger("speaker_classifier")
            self.logger.setLevel(logging.INFO)
            handler = logging.FileHandler(self.log_file, "w", "utf-8")
            handler.setFormatter = logging.Formatter("%(name)s %(message)s")
            self.logger.addHandler(handler)
        else:
            self.logger = logger
        self.classifier = None
        self.speaker_labels = {}
        self.ivectors = {}
        self.num_speakers = num_speakers
        self.cluster = cluster
        self.uses_voiced = False
        self.uses_cmvn = True
        self.uses_splices = False
        self.setup()

    @property
    def classify_directory(self) -> str:
        """Temporary directory for speaker classification"""
        return os.path.join(self.temp_directory, "speaker_classification")

    @property
    def data_directory(self) -> str:
        """Corpus data directory"""
        return self.corpus.split_directory

    @property
    def working_directory(self) -> str:
        """Current working directory for the speaker classifier"""
        return self.classify_directory

    @property
    def ie_path(self) -> str:
        """Path for the IvectorExtractor model file"""
        return os.path.join(self.working_directory, "final.ie")

    @property
    def speaker_classification_model_path(self) -> str:
        """Path for the speaker classification model"""
        return os.path.join(self.working_directory, "speaker_classifier.mdl")

    @property
    def speaker_labels_path(self) -> str:
        """Path for the speaker labels file"""
        return os.path.join(self.working_directory, "speaker_labels.txt")

    @property
    def model_path(self) -> str:
        """Path for the acoustic model file"""
        return os.path.join(self.working_directory, "final.mdl")

    @property
    def dubm_path(self) -> str:
        """Path for the DUBM model"""
        return os.path.join(self.working_directory, "final.dubm")

    @property
    def working_log_directory(self) -> str:
        """Current log directory"""
        return self.log_directory

    @property
    def ivector_options(self) -> MetaDict:
        """Ivector configuration options"""
        data = self.ivector_extractor.meta
        data["silence_weight"] = 0.0
        data["posterior_scale"] = 0.1
        data["max_count"] = 100
        data["sil_phones"] = None
        return data

    @property
    def log_directory(self) -> str:
        """Log directory"""
        return os.path.join(self.classify_directory, "log")

    @property
    def use_mp(self) -> bool:
        """Flag for whether to use multiprocessing"""
        return self.classification_config.use_mp

    def setup(self) -> None:
        """
        Sets up the corpus and speaker classifier

        Raises
        ------
        :class:`~montreal_forced_aligner.exceptions.KaldiProcessingError`
            If there were any errors in running Kaldi binaries
        """
        done_path = os.path.join(self.classify_directory, "done")
        if os.path.exists(done_path):
            self.logger.info("Classification already done, skipping initialization.")
            return
        dirty_path = os.path.join(self.classify_directory, "dirty")
        if os.path.exists(dirty_path):
            shutil.rmtree(self.classify_directory)
        log_dir = os.path.join(self.classify_directory, "log")
        os.makedirs(log_dir, exist_ok=True)
        self.ivector_extractor.export_model(self.classify_directory)
        try:
            self.corpus.initialize_corpus(None, self.feature_config)
            extract_ivectors(self)
        except Exception as e:
            with open(dirty_path, "w"):
                pass
            if isinstance(e, KaldiProcessingError):
                log_kaldi_errors(e.error_logs, self.logger)
                e.update_log_file(self.logger.handlers[0].baseFilename)
            raise

    def load_ivectors(self) -> None:
        """
        Load ivectors from the temporary directory
        """
        self.ivectors = {}
        for j in self.corpus.jobs:
            ivectors_args = j.extract_ivector_arguments(self)
            for ivectors_path in ivectors_args.ivector_paths.values():
                ivec = load_scp(ivectors_path)
                for utt, ivector in ivec.items():
                    ivector = [float(x) for x in ivector]
                    self.ivectors[utt] = ivector

    def cluster_utterances(self) -> None:
        """
        Cluster utterances based on their ivectors
        """
        from sklearn.cluster import KMeans

        if not self.ivectors:
            self.load_ivectors()
        x = []
        for v in self.ivectors.values():
            x.append(v)
        x = np.array(x)
        clust = KMeans(self.num_speakers).fit(x)
        y = clust.labels_
        for i, u in enumerate(self.ivectors.keys()):
            speaker_name = y[i]
            utterance = self.corpus.utterances[u]
            if speaker_name not in self.corpus.speakers:
                self.corpus.speakers[speaker_name] = Speaker(speaker_name)
            utterance.set_speaker(self.corpus.speakers[speaker_name])

    def export_classification(self, output_directory: str) -> None:
        """
        Export files with their new speaker labels

        Parameters
        ----------
        output_directory: str
            Output directory to save files
        """
        backup_output_directory = None
        if not self.classification_config.overwrite:
            backup_output_directory = os.path.join(self.classify_directory, "output")
            os.makedirs(backup_output_directory, exist_ok=True)

        for file in self.corpus.files.values():
            file.save(output_directory, backup_output_directory)



"""Class definitions for Montreal Forced Aligner models"""
from __future__ import annotations

import os
from shutil import copy, copyfile, make_archive, move, rmtree, unpack_archive
from typing import TYPE_CHECKING, Any, Collection, Dict, Optional, Union

import yaml

from .exceptions import (
    LanguageModelNotFoundError,
    ModelLoadError,
    PronunciationAcousticMismatchError,
)
from .helper import TerminalPrinter

if TYPE_CHECKING:
    from logging import Logger

    from .aligner.adapting import AdaptingAligner
    from .config import FeatureConfig
    from .config.train_config import TrainingConfig
    from .dictionary import Dictionary
    from .lm.trainer import LmTrainer
    from .trainers import BaseTrainer

    TrainerType = Union[BaseTrainer, LmTrainer, AdaptingAligner]
    MetaDict = Dict[str, Any]


# default format for output
FORMAT = "zip"

__all__ = [
    "Archive",
    "LanguageModel",
    "AcousticModel",
    "IvectorExtractor",
    "DictionaryModel",
    "G2PModel",
    "MODEL_TYPES",
]


class Archive:
    """
    Class representing data in a directory or archive file (zip, tar,
    tar.gz/tgz)

    Based on the prosodylab-aligner
    (https://github.com/prosodylab/Prosodylab-Aligner) archive class.

    Parameters
    ----------
    source: str
        Source path
    root_directory: str
        Root directory to unpack and store temporary files
    """

    extensions = [".zip"]

    def __init__(self, source: str, root_directory: Optional[str] = None):
        from .config import TEMP_DIR

        if root_directory is None:
            root_directory = TEMP_DIR
        self.root_directory = root_directory
        self.source = source
        self._meta = {}
        self.name, _ = os.path.splitext(os.path.basename(source))
        if os.path.isdir(source):
            self.dirname = os.path.abspath(source)
        else:
            self.dirname = os.path.join(root_directory, self.name)
            if not os.path.exists(self.dirname):
                os.makedirs(root_directory, exist_ok=True)
                unpack_archive(source, self.dirname)
                files = os.listdir(self.dirname)
                old_dir_path = os.path.join(self.dirname, files[0])
                if len(files) == 1 and os.path.isdir(old_dir_path):  # Backwards compatibility
                    for f in os.listdir(old_dir_path):
                        move(os.path.join(old_dir_path, f), os.path.join(self.dirname, f))
                    os.rmdir(old_dir_path)

    def get_subclass_object(
        self,
    ) -> Union[AcousticModel, G2PModel, LanguageModel, IvectorExtractor]:
        """
        Instantiate subclass models based on files contained in the archive

        Returns
        -------
        Union[AcousticModel, G2PModel, LanguageModel, IvectorExtractor]
            Subclass model that was auto detected
        """
        for f in os.listdir(self.dirname):
            if f == "tree":
                return AcousticModel(self.dirname, self.root_directory)
            if f == "phones.sym":
                return G2PModel(self.dirname, self.root_directory)
            if f.endswith(".arpa"):
                return LanguageModel(self.dirname, self.root_directory)
            if f == "final.ie":
                return IvectorExtractor(self.dirname, self.root_directory)
        raise ModelLoadError(self.source)

    @classmethod
    def valid_extension(cls, filename: str) -> bool:
        """
        Check whether a file has a valid extension for the given model archive

        Parameters
        ----------
        filename: str
            File name to check

        Returns
        -------
        bool
            True if the extension matches the models allowed extensions
        """
        if os.path.splitext(filename)[1] in cls.extensions:
            return True
        return False

    @classmethod
    def generate_path(cls, root: str, name: str, enforce_existence: bool = True) -> Optional[str]:
        """
        Generate a path for a given model from the root directory and the name of the model

        Parameters
        ----------
        root: str
            Root directory for the full path
        name: str
            Name of the model
        enforce_existence: bool
            Flag to return None if the path doesn't exist, defaults to True

        Returns
        -------
        str
           Full path in the root directory for the model
        """
        for ext in cls.extensions:
            path = os.path.join(root, name + ext)
            if os.path.exists(path) or not enforce_existence:
                return path
        return None

    def pretty_print(self):
        """
        Pretty print the archive's meta data using TerminalPrinter
        """
        printer = TerminalPrinter()
        configuration_data = {"Archive": {"name": (self.name, "green"), "data": self.meta}}
        printer.print_config(configuration_data)

    @property
    def meta(self) -> dict:
        """
        Get the meta data associated with the model
        """
        if not self._meta:
            meta_path = os.path.join(self.dirname, "meta.yaml")
            with open(meta_path, "r", encoding="utf8") as f:
                self._meta = yaml.safe_load(f)
        return self._meta

    def add_meta_file(self, trainer: TrainerType) -> None:
        """
        Add a metadata file from a given trainer to the model

        Parameters
        ----------
        trainer: TrainerType
            The trainer to construct the metadata from
        """
        with open(os.path.join(self.dirname, "meta.yaml"), "w", encoding="utf8") as f:
            yaml.dump(trainer.meta, f)

    @classmethod
    def empty(cls, head: str, root_directory: Optional[str] = None) -> Archive:
        """
        Initialize an archive using an empty directory

        Parameters
        ----------
        head: str
            Directory name to create
        root_directory: str, optional
            Root directory to create temporary data, defaults to the MFA temporary directory

        Returns
        -------
        Archive
            Model constructed from the empty directory
        """
        from .config import TEMP_DIR

        if root_directory is None:
            root_directory = TEMP_DIR

        os.makedirs(root_directory, exist_ok=True)
        source = os.path.join(root_directory, head)
        os.makedirs(source, exist_ok=True)
        return cls(source)

    def add(self, source: str):
        """
        Add file into archive

        Parameters
        ----------
        source: str
            Path to file to copy into the directory
        """
        copy(source, self.dirname)

    def __repr__(self) -> str:
        """Representation string of a model"""
        return f"{self.__class__.__name__}(dirname={self.dirname!r})"

    def clean_up(self) -> None:
        """Remove temporary directory"""
        rmtree(self.dirname)

    def dump(self, path: str, archive_fmt: str = FORMAT) -> str:
        """
        Write archive to disk, and return the name of final archive

        Parameters
        ----------
        path: str
            Path to write to
        archive_fmt: str, optional
            Archive extension to use, defaults to ".zip"

        Returns
        -------
        str
            Path of constructed archive
        """
        return make_archive(os.path.splitext(path)[0], archive_fmt, *os.path.split(self.dirname))


class AcousticModel(Archive):
    """
    Class for storing acoustic models in MFA, exported as zip files containing the necessary Kaldi files
    to be reused

    """

    files = ["final.mdl", "final.alimdl", "final.occs", "lda.mat", "tree"]
    extensions = [".zip", ".am"]

    def add_meta_file(self, aligner: TrainerType) -> None:
        """
        Add metadata file from a model trainer

        Parameters
        ----------
        aligner: TrainerType
            Trainer to supply metadata information about the acoustic model
        """
        with open(os.path.join(self.dirname, "meta.yaml"), "w", encoding="utf8") as f:
            yaml.dump(aligner.meta, f)

    @property
    def feature_config(self) -> FeatureConfig:
        """
        Return the FeatureConfig used in training the model
        """
        from .config.feature_config import FeatureConfig

        fc = FeatureConfig()
        fc.update(self.meta["features"])
        return fc

    def adaptation_config(self) -> TrainingConfig:
        """
        Generate an adaptation configuration

        Returns
        -------
        TrainingConfig
            Configuration to be used in adapting the acoustic model to new data
        """
        from .config.train_config import load_no_sat_adapt, load_sat_adapt

        if self.meta["features"]["fmllr"]:
            train, align = load_sat_adapt()
        else:
            train, align = load_no_sat_adapt()
        return train

    @property
    def meta(self) -> MetaDict:
        """
        Metadata information for the acoustic model
        """
        default_features = {
            "type": "mfcc",
            "use_energy": False,
            "frame_shift": 10,
            "pitch": False,
            "fmllr": True,
        }
        if not self._meta:
            meta_path = os.path.join(self.dirname, "meta.yaml")
            if not os.path.exists(meta_path):
                self._meta = {
                    "version": "0.9.0",
                    "architecture": "gmm-hmm",
                    "multilingual_ipa": False,
                    "features": default_features,
                }
            else:
                with open(meta_path, "r", encoding="utf8") as f:
                    self._meta = yaml.safe_load(f)
                if self._meta["features"] == "mfcc+deltas":
                    self._meta["features"] = default_features
            if "uses_lda" not in self._meta:  # Backwards compatibility
                self._meta["uses_lda"] = os.path.exists(os.path.join(self.dirname, "lda.mat"))
            if "multilingual_ipa" not in self._meta:
                self._meta["multilingual_ipa"] = False
            if "uses_sat" not in self._meta:
                self._meta["uses_sat"] = False
            if "phone_type" not in self._meta:
                self._meta["phone_type"] = "triphone"
            self._meta["phones"] = set(self._meta.get("phones", []))
            self._meta["has_speaker_independent_model"] = os.path.exists(
                os.path.join(self.dirname, "final.alimdl")
            )
        return self._meta

    def pretty_print(self) -> None:
        """
        Prints the metadata information to the terminal
        """
        from .utils import get_mfa_version

        printer = TerminalPrinter()
        configuration_data = {"Acoustic model": {"name": (self.name, "green"), "data": {}}}
        version_color = "green"
        if self.meta["version"] != get_mfa_version():
            version_color = "red"
        configuration_data["Acoustic model"]["data"]["Version"] = (
            self.meta["version"],
            version_color,
        )

        if "citation" in self.meta:
            configuration_data["Acoustic model"]["data"]["Citation"] = self.meta["citation"]
        if "train_date" in self.meta:
            configuration_data["Acoustic model"]["data"]["Train date"] = self.meta["train_date"]
        configuration_data["Acoustic model"]["data"]["Architecture"] = self.meta["architecture"]
        configuration_data["Acoustic model"]["data"]["Phone type"] = self.meta["phone_type"]
        configuration_data["Acoustic model"]["data"]["Features"] = {
            "Type": self.meta["features"]["type"],
            "Frame shift": self.meta["features"]["frame_shift"],
        }
        if self.meta["phones"]:
            configuration_data["Acoustic model"]["data"]["Phones"] = self.meta["phones"]
        else:
            configuration_data["Acoustic model"]["data"]["Phones"] = ("None found!", "red")

        configuration_data["Acoustic model"]["data"]["Configuration options"] = {
            "Multilingual IPA": self.meta["multilingual_ipa"],
            "Performs speaker adaptation": self.meta["uses_sat"],
            "Has speaker-independent model": self.meta["has_speaker_independent_model"],
            "Performs LDA on features": self.meta["uses_lda"],
        }
        printer.print_config(configuration_data)

    def add_model(self, source: str) -> None:
        """
        Add file into archive

        Parameters
        ----------
        source: str
            File to add
        """
        for f in self.files:
            if os.path.exists(os.path.join(source, f)):
                copyfile(os.path.join(source, f), os.path.join(self.dirname, f))

    def export_model(self, destination: str) -> None:
        """
        Extract the model files to a new directory

        Parameters
        ----------
        destination: str
            Destination directory to extract files to
        """
        os.makedirs(destination, exist_ok=True)
        for f in self.files:
            if os.path.exists(os.path.join(self.dirname, f)):
                copyfile(os.path.join(self.dirname, f), os.path.join(destination, f))

    def log_details(self, logger: Logger) -> None:
        """
        Log metadata information to a logger

        Parameters
        ----------
        logger: :class:`~logging.Logger`
            Logger to send debug information to
        """
        logger.debug("")
        logger.debug("====ACOUSTIC MODEL INFO====")
        logger.debug("Acoustic model root directory: " + self.root_directory)
        logger.debug("Acoustic model dirname: " + self.dirname)
        meta_path = os.path.join(self.dirname, "meta.yaml")
        logger.debug("Acoustic model meta path: " + meta_path)
        if not os.path.exists(meta_path):
            logger.debug("META.YAML DOES NOT EXIST, this may cause issues in validating the model")
        logger.debug("Acoustic model meta information:")
        stream = yaml.dump(self.meta)
        logger.debug(stream)
        logger.debug("")

    def validate(self, dictionary: Union[Dictionary, G2PModel]) -> None:
        """
        Validate this acoustic model against a pronunciation dictionary or G2P model to ensure their
        phone sets are compatible

        Parameters
        ----------
        dictionary: Union[Dictionary, G2PModel]
            Dictionary or G2P model to compare phone sets with

        Raises
        ------
        PronunciationAcousticMismatchError
            If there are phones missing from the acoustic model
        """
        if isinstance(dictionary, G2PModel):
            missing_phones = dictionary.meta["phones"] - set(self.meta["phones"])
        else:
            missing_phones = dictionary.nonsil_phones - set(self.meta["phones"])
        if missing_phones:
            raise (PronunciationAcousticMismatchError(missing_phones))


class IvectorExtractor(Archive):
    """
    Model class for IvectorExtractor
    """

    model_files = [
        "final.ie",
        "final.ubm",
        "final.dubm",
        "plda",
        "mean.vec",
        "trans.mat",
    ]
    extensions = [".zip", ".ivector"]

    def add_model(self, source: str) -> None:
        """
        Add file into archive

        Parameters
        ----------
        source: str
            File to add
        """
        for filename in self.model_files:
            if os.path.exists(os.path.join(source, filename)):
                copyfile(os.path.join(source, filename), os.path.join(self.dirname, filename))

    def export_model(self, destination: str) -> None:
        """
        Extract the model files to a new directory

        Parameters
        ----------
        destination: str
            Destination directory to extract files to
        """
        os.makedirs(destination, exist_ok=True)
        for filename in self.model_files:
            if os.path.exists(os.path.join(self.dirname, filename)):
                copyfile(os.path.join(self.dirname, filename), os.path.join(destination, filename))

    @property
    def feature_config(self) -> FeatureConfig:
        """
        Return the FeatureConfig used in training the model
        """
        from .config.feature_config import FeatureConfig

        fc = FeatureConfig()
        fc.update(self.meta["features"])
        return fc


class G2PModel(Archive):
    extensions = [".zip", ".g2p"]

    def add_meta_file(self, dictionary: Dictionary, architecture: Optional[str] = None) -> None:
        """
        Construct meta data information for the G2P model from the dictionary it was trained from

        Parameters
        ----------
        dictionary: Dictionary
            Dictionary that was the training data for the G2P model
        architecture: str, optional
            Architecture of the G2P model, defaults to "pynini"
        """
        from .utils import get_mfa_version

        if architecture is None:
            architecture = "pynini"
        with open(os.path.join(self.dirname, "meta.yaml"), "w", encoding="utf8") as f:
            meta = {
                "phones": sorted(dictionary.nonsil_phones),
                "graphemes": sorted(dictionary.graphemes),
                "architecture": architecture,
                "version": get_mfa_version(),
            }
            yaml.dump(meta, f)

    @property
    def meta(self) -> dict:
        """Metadata for the G2P model"""
        if not self._meta:
            meta_path = os.path.join(self.dirname, "meta.yaml")
            if not os.path.exists(meta_path):
                self._meta = {"version": "0.9.0", "architecture": "phonetisaurus"}
            else:
                with open(meta_path, "r", encoding="utf8") as f:
                    self._meta = yaml.safe_load(f)
            self._meta["phones"] = set(self._meta.get("phones", []))
            self._meta["graphemes"] = set(self._meta.get("graphemes", []))
        return self._meta

    @property
    def fst_path(self) -> str:
        """G2P model's FST path"""
        return os.path.join(self.dirname, "model.fst")

    @property
    def sym_path(self) -> str:
        """G2p model's symbols path"""
        return os.path.join(self.dirname, "phones.sym")

    def add_sym_path(self, source_directory: str) -> None:
        """
        Add symbols file into archive

        Parameters
        ----------
        source: str
            Source directory path
        """
        if not os.path.exists(self.sym_path):
            copyfile(os.path.join(source_directory, "phones.sym"), self.sym_path)

    def add_fst_model(self, source_directory: str) -> None:
        """
        Add FST file into archive

        Parameters
        ----------
        source: str
            Source directory path
        """
        if not os.path.exists(self.fst_path):
            copyfile(os.path.join(source_directory, "model.fst"), self.fst_path)

    def export_fst_model(self, destination: str) -> None:
        """
        Extract FST model path to destination

        Parameters
        ----------
        destination: str
            Destination directory
        """
        os.makedirs(destination, exist_ok=True)
        copy(self.fst_path, destination)

    def validate(self, word_list: Collection[str]) -> bool:
        """
        Validate the G2P model against a word list to ensure that all graphemes are known

        Parameters
        ----------
        word_list: Collection[str]
            Word list to validate against

        Returns
        -------
        bool
            False if missing graphemes were found
        """
        graphemes = set()
        for w in word_list:
            graphemes.update(w)
        missing_graphemes = graphemes - self.meta["graphemes"]
        if missing_graphemes:
            print(
                "WARNING! The following graphemes were not found in the specified G2P model: "
                f"{' '.join(sorted(missing_graphemes))}"
            )
            return False
        else:
            return True


class LanguageModel(Archive):
    """
    Class for MFA language models
    """

    arpa_extension = ".arpa"
    extensions = [f".{FORMAT}", arpa_extension, ".lm"]

    def __init__(self, source: str, root_directory: Optional[str] = None):
        from .config import TEMP_DIR

        if root_directory is None:
            root_directory = TEMP_DIR
        self.root_directory = root_directory
        self._meta = {}
        self.name, _ = os.path.splitext(os.path.basename(source))
        if os.path.isdir(source):
            self.dirname = os.path.abspath(source)
        elif source.endswith(self.arpa_extension):
            self.dirname = os.path.join(root_directory, self.name)
            if not os.path.exists(self.dirname):
                os.makedirs(self.dirname, exist_ok=True)
            copy(source, self.large_arpa_path)
        elif any(source.endswith(x) for x in self.extensions):
            base = root_directory
            self.dirname = os.path.join(root_directory, self.name)
            if not os.path.exists(self.dirname):
                os.makedirs(root_directory, exist_ok=True)
                unpack_archive(source, base)

    @property
    def decode_arpa_path(self) -> str:
        """
        Uses the smallest language model for decoding
        """
        for path in [self.small_arpa_path, self.medium_arpa_path, self.large_arpa_path]:
            if os.path.exists(path):
                return path
        raise LanguageModelNotFoundError()

    @property
    def carpa_path(self) -> str:
        """
        Uses the largest language model for rescoring
        """
        for path in [self.large_arpa_path, self.medium_arpa_path, self.small_arpa_path]:
            if os.path.exists(path):
                return path
        raise LanguageModelNotFoundError()

    @property
    def small_arpa_path(self) -> str:
        """Small arpa path"""
        return os.path.join(self.dirname, self.name + "_small" + self.arpa_extension)

    @property
    def medium_arpa_path(self) -> str:
        """Medium arpa path"""
        return os.path.join(self.dirname, self.name + "_med" + self.arpa_extension)

    @property
    def large_arpa_path(self) -> str:
        """Large arpa path"""
        return os.path.join(self.dirname, self.name + self.arpa_extension)

    def add_arpa_file(self, arpa_path: str) -> None:
        """
        Adds an ARPA file to the model

        Parameters
        ----------
        arpa_path: str
            Path to ARPA file
        """
        name = os.path.basename(arpa_path)
        copyfile(arpa_path, os.path.join(self.dirname, name))


class DictionaryModel(Archive):
    """
    Class for representing MFA pronunciation dictionaries
    """

    extensions = [".dict", f".{FORMAT}", ".txt", ".yaml"]


MODEL_TYPES = {
    "acoustic": AcousticModel,
    "g2p": G2PModel,
    "dictionary": DictionaryModel,
    "language_model": LanguageModel,
    "ivector": IvectorExtractor,
}



"""Class definitions for TriphoneTrainer"""
from __future__ import annotations

import os
import subprocess
import time
from typing import TYPE_CHECKING, Optional

from ..exceptions import KaldiProcessingError
from ..multiprocessing import compile_train_graphs, convert_alignments, tree_stats
from ..utils import log_kaldi_errors, parse_logs, thirdparty_binary
from .base import BaseTrainer

if TYPE_CHECKING:
    from ..config import FeatureConfig
    from ..corpus import Corpus
    from ..dictionary import DictionaryType
    from .base import TrainerType


__all__ = ["TriphoneTrainer"]


class TriphoneTrainer(BaseTrainer):
    """
    Configuration class for triphone training

    Attributes
    ----------
    num_iterations : int
        Number of training iterations to perform, defaults to 40
    max_gaussians : int
        Total number of gaussians, defaults to 1000
    num_leaves : int
        Number of states in the decision tree, defaults to 1000
    max_gaussians : int
        Number of gaussians in the decision tree, defaults to 10000
    cluster_threshold : int
        For build-tree control final bottom-up clustering of leaves, defaults to 100
    """

    def __init__(self, default_feature_config: FeatureConfig):
        super(TriphoneTrainer, self).__init__(default_feature_config)

        self.num_iterations = 35
        self.num_leaves = 1000
        self.max_gaussians = 10000
        self.cluster_threshold = -1
        self.compute_calculated_properties()

    def compute_calculated_properties(self) -> None:
        """Generate realignment iterations and initial gaussians based on configuration"""
        for i in range(0, self.num_iterations, 10):
            if i == 0:
                continue
            self.realignment_iterations.append(i)
        self.initial_gaussians = self.num_leaves
        self.current_gaussians = self.num_leaves

    @property
    def train_type(self) -> str:
        """Training identifier"""
        return "tri"

    @property
    def phone_type(self) -> str:
        """Phone type"""
        return "triphone"

    def _setup_tree(self) -> None:
        """
        Set up the tree for the triphone model

        Raises
        ------
        :class:`~montreal_forced_aligner.exceptions.KaldiProcessingError`
            If there were any errors in running Kaldi binaries
        """
        dirty_path = os.path.join(self.train_directory, "dirty")
        try:

            tree_stats(self)
            log_path = os.path.join(self.log_directory, "questions.log")
            tree_path = os.path.join(self.train_directory, "tree")
            treeacc_path = os.path.join(self.train_directory, "treeacc")
            sets_int_path = os.path.join(self.dictionary.phones_dir, "sets.int")
            roots_int_path = os.path.join(self.dictionary.phones_dir, "roots.int")
            extra_question_int_path = os.path.join(
                self.dictionary.phones_dir, "extra_questions.int"
            )
            topo_path = os.path.join(self.dictionary.output_directory, "topo")
            questions_path = os.path.join(self.train_directory, "questions.int")
            questions_qst_path = os.path.join(self.train_directory, "questions.qst")
            with open(log_path, "w") as log_file:
                subprocess.call(
                    [
                        thirdparty_binary("cluster-phones"),
                        treeacc_path,
                        sets_int_path,
                        questions_path,
                    ],
                    stderr=log_file,
                )

            with open(extra_question_int_path, "r") as inf, open(questions_path, "a") as outf:
                for line in inf:
                    outf.write(line)

            log_path = os.path.join(self.log_directory, "compile_questions.log")
            with open(log_path, "w") as log_file:
                subprocess.call(
                    [
                        thirdparty_binary("compile-questions"),
                        topo_path,
                        questions_path,
                        questions_qst_path,
                    ],
                    stderr=log_file,
                )

            log_path = os.path.join(self.log_directory, "build_tree.log")
            with open(log_path, "w") as log_file:
                subprocess.call(
                    [
                        thirdparty_binary("build-tree"),
                        "--verbose=1",
                        f"--max-leaves={self.initial_gaussians}",
                        f"--cluster-thresh={self.cluster_threshold}",
                        treeacc_path,
                        roots_int_path,
                        questions_qst_path,
                        topo_path,
                        tree_path,
                    ],
                    stderr=log_file,
                )

            log_path = os.path.join(self.log_directory, "init_model.log")
            occs_path = os.path.join(self.train_directory, "0.occs")
            mdl_path = self.current_model_path
            with open(log_path, "w") as log_file:
                subprocess.call(
                    [
                        thirdparty_binary("gmm-init-model"),
                        f"--write-occs={occs_path}",
                        tree_path,
                        treeacc_path,
                        topo_path,
                        mdl_path,
                    ],
                    stderr=log_file,
                )

            log_path = os.path.join(self.log_directory, "mixup.log")
            with open(log_path, "w") as log_file:
                subprocess.call(
                    [
                        thirdparty_binary("gmm-mixup"),
                        f"--mix-up={self.initial_gaussians}",
                        mdl_path,
                        occs_path,
                        mdl_path,
                    ],
                    stderr=log_file,
                )
            os.remove(treeacc_path)
            parse_logs(self.log_directory)

            compile_train_graphs(self)

            convert_alignments(self)
            os.rename(occs_path, self.next_occs_path)
            os.rename(mdl_path, self.next_model_path)
        except Exception as e:
            with open(dirty_path, "w"):
                pass
            if isinstance(e, KaldiProcessingError):
                log_kaldi_errors(e.error_logs, self.logger)
                e.update_log_file(self.logger.handlers[0].baseFilename)
            raise

    def init_training(
        self,
        identifier: str,
        temporary_directory: str,
        corpus: Corpus,
        dictionary: DictionaryType,
        previous_trainer: Optional[TrainerType],
    ):
        """
        Initialize triphone training

        Parameters
        ----------
        identifier: str
            Identifier for the training block
        temporary_directory: str
            Root temporary directory to save
        corpus: :class:`~montreal_forced_aligner.corpus.base.Corpus`
            Corpus to use
        dictionary: DictionaryType
            Dictionary to use
        previous_trainer: TrainerType, optional
            Previous trainer to initialize from
        """
        self._setup_for_init(identifier, temporary_directory, corpus, dictionary, previous_trainer)
        done_path = os.path.join(self.train_directory, "done")
        if os.path.exists(done_path):
            self.logger.info(f"{self.identifier} training already done, skipping initialization.")
            return
        begin = time.time()
        self._setup_tree()

        self.iteration = 1
        self.logger.info("Initialization complete!")
        self.logger.debug(f"Initialization took {time.time() - begin} seconds")



"""Class definitions for configuring speaker classification"""
from __future__ import annotations

import os

import yaml

from .base_config import BaseConfig

__all__ = [
    "SpeakerClassificationConfig",
    "classification_yaml_to_config",
    "load_basic_classification",
]


class SpeakerClassificationConfig(BaseConfig):
    """
    Configuration class to store parameters for speaker classification
    """

    def __init__(self):
        self.use_mp = True
        self.pca_dimension = -1
        self.target_energy = 0.1
        self.cluster_threshold = 0.5
        self.max_speaker_fraction = 1.0
        self.first_pass_max_utterances = 32767
        self.rttm_channel = 0
        self.read_costs = False
        self.overwrite = False


def classification_yaml_to_config(path: str) -> SpeakerClassificationConfig:
    """
    Helper function to load speaker classification configurations

    Parameters
    ----------
    path: str
        Path to yaml file

    Returns
    -------
    :class:`~montreal_forced_aligner.config.speaker_classification_config.SpeakerClassificationConfig`
        Speaker classification configuration
    """
    with open(path, "r", encoding="utf8") as f:
        data = yaml.load(f, Loader=yaml.SafeLoader)
        classification_config = SpeakerClassificationConfig()
        if data:
            classification_config.update(data)
        return classification_config


def load_basic_classification() -> SpeakerClassificationConfig:
    """
    Helper function to load the default parameters

    Returns
    -------
    :class:`~montreal_forced_aligner.config.speaker_classification_config.SpeakerClassificationConfig`
        Default speaker classification configuration
    """
    base_dir = os.path.dirname(os.path.abspath(__file__))
    classification_config = classification_yaml_to_config(
        os.path.join(base_dir, "basic_classification.yaml")
    )
    return classification_config



"""Command line functions for segmenting audio files"""
from __future__ import annotations

import os
import shutil
import time
from typing import TYPE_CHECKING, Optional

from montreal_forced_aligner.config import (
    TEMP_DIR,
    load_basic_segmentation,
    load_command_configuration,
    segmentation_yaml_to_config,
)
from montreal_forced_aligner.corpus import Corpus
from montreal_forced_aligner.exceptions import ArgumentError
from montreal_forced_aligner.segmenter import Segmenter
from montreal_forced_aligner.utils import log_config, setup_logger

if TYPE_CHECKING:
    from argparse import Namespace


__all__ = ["create_segments", "validate_args", "run_create_segments"]


def create_segments(args: Namespace, unknown_args: Optional[list] = None) -> None:
    """
    Run the sound file segmentation

    Parameters
    ----------
    args: :class:`~argparse.Namespace`
        Command line arguments
    unknown_args: List[str]
        Optional arguments that will be passed to configuration objects
    """
    from montreal_forced_aligner.utils import get_mfa_version

    command = "create_segments"
    all_begin = time.time()
    if not args.temp_directory:
        temp_dir = TEMP_DIR
    else:
        temp_dir = os.path.expanduser(args.temp_directory)
    corpus_name = os.path.basename(args.corpus_directory)
    if corpus_name == "":
        args.corpus_directory = os.path.dirname(args.corpus_directory)
        corpus_name = os.path.basename(args.corpus_directory)
    data_directory = os.path.join(temp_dir, corpus_name)
    conf_path = os.path.join(data_directory, "config.yml")
    if args.config_path:
        segmentation_config = segmentation_yaml_to_config(args.config_path)
    else:
        segmentation_config = load_basic_segmentation()
    segmentation_config.update_from_args(args)
    if unknown_args:
        segmentation_config.update_from_unknown_args(unknown_args)
    if getattr(args, "clean", False) and os.path.exists(data_directory):
        print("Cleaning old directory!")
        shutil.rmtree(data_directory, ignore_errors=True)
    if getattr(args, "verbose", False):
        log_level = "debug"
    else:
        log_level = "info"
    logger = setup_logger(command, data_directory, console_level=log_level)
    log_config(logger, segmentation_config)
    conf = load_command_configuration(
        conf_path,
        {
            "dirty": False,
            "begin": time.time(),
            "version": get_mfa_version(),
            "type": command,
            "corpus_directory": args.corpus_directory,
        },
    )
    if (
        conf["dirty"]
        or conf["type"] != command
        or conf["corpus_directory"] != args.corpus_directory
        or conf["version"] != get_mfa_version()
    ):
        logger.warning(
            "WARNING: Using old temp directory, this might not be ideal for you, use the --clean flag to ensure no "
            "weird behavior for previous versions of the temporary directory."
        )
        if conf["dirty"]:
            logger.debug("Previous run ended in an error (maybe ctrl-c?)")
        if conf["type"] != command:
            logger.debug(
                f"Previous run was a different subcommand than {command} (was {conf['type']})"
            )
        if conf["corpus_directory"] != args.corpus_directory:
            logger.debug(
                "Previous run used source directory "
                f"path {conf['corpus_directory']} (new run: {args.corpus_directory})"
            )
        if conf["version"] != get_mfa_version():
            logger.debug(
                f"Previous run was on {conf['version']} version (new run: {get_mfa_version()})"
            )

    os.makedirs(data_directory, exist_ok=True)
    os.makedirs(args.output_directory, exist_ok=True)
    try:
        corpus = Corpus(
            args.corpus_directory,
            data_directory,
            sample_rate=segmentation_config.feature_config.sample_frequency,
            num_jobs=args.num_jobs,
            logger=logger,
            use_mp=segmentation_config.use_mp,
            ignore_speakers=True,
        )

        begin = time.time()
        a = Segmenter(
            corpus,
            segmentation_config,
            temp_directory=data_directory,
            debug=getattr(args, "debug", False),
            logger=logger,
        )
        logger.debug(f"Setup segmenter in {time.time() - begin} seconds")
        a.verbose = args.verbose

        begin = time.time()
        a.segment()
        logger.debug(f"Performed segmentation in {time.time() - begin} seconds")

        begin = time.time()
        a.export_segments(args.output_directory)
        logger.debug(f"Exported segmentation in {time.time() - begin} seconds")
        logger.info("Done!")
        logger.debug(f"Done! Everything took {time.time() - all_begin} seconds")
    except Exception:
        conf["dirty"] = True
        raise
    finally:
        handlers = logger.handlers[:]
        for handler in handlers:
            handler.close()
            logger.removeHandler(handler)
        conf.save(conf_path)


def validate_args(args: Namespace) -> None:
    """
    Validate the command line arguments

    Parameters
    ----------
    args: :class:`~argparse.Namespace`
        Parsed command line arguments

    Raises
    ------
    :class:`~montreal_forced_aligner.exceptions.ArgumentError`
        If there is a problem with any arguments
    """
    args.output_directory = args.output_directory.rstrip("/").rstrip("\\")
    args.corpus_directory = args.corpus_directory.rstrip("/").rstrip("\\")
    if not os.path.exists(args.corpus_directory):
        raise ArgumentError(f"Could not find the corpus directory {args.corpus_directory}.")
    if not os.path.isdir(args.corpus_directory):
        raise ArgumentError(
            f"The specified corpus directory ({args.corpus_directory}) is not a directory."
        )

    if args.corpus_directory == args.output_directory:
        raise ArgumentError("Corpus directory and output directory cannot be the same folder.")


def run_create_segments(args: Namespace, unknown: Optional[list] = None) -> None:
    """
    Wrapper function for running sound file segmentation

    Parameters
    ----------
    args: :class:`~argparse.Namespace`
        Parsed command line arguments
    unknown: List[str]
        Parsed command line arguments to be passed to the configuration objects
    """
    validate_args(args)
    create_segments(args, unknown)



"""Class definitions for LDA trainer"""
from __future__ import annotations

import os
import time
from typing import TYPE_CHECKING, Optional

from ..exceptions import KaldiProcessingError
from ..multiprocessing import (
    acc_stats,
    align,
    calc_lda_mllt,
    compute_alignment_improvement,
    lda_acc_stats,
)
from ..utils import log_kaldi_errors, parse_logs
from .triphone import TriphoneTrainer

if TYPE_CHECKING:
    from ..config import FeatureConfig
    from ..corpus import Corpus
    from ..dictionary import DictionaryType
    from .base import MetaDict, TrainerType


__all__ = ["LdaTrainer"]


class LdaTrainer(TriphoneTrainer):
    """

    Configuration class for LDA+MLLT training

    Attributes
    ----------
    lda_dimension : int
        Dimensionality of the LDA matrix
    mllt_iterations : list
        List of iterations to perform MLLT estimation
    random_prune : float
        This is approximately the ratio by which we will speed up the
        LDA and MLLT calculations via randomized pruning
    """

    def __init__(self, default_feature_config: FeatureConfig):
        super(LdaTrainer, self).__init__(default_feature_config)
        self.lda_dimension = 40
        self.mllt_iterations = []
        max_mllt_iter = int(self.num_iterations / 2) - 1
        for i in range(1, max_mllt_iter):
            if i < max_mllt_iter / 2 and i % 2 == 0:
                self.mllt_iterations.append(i)
        self.mllt_iterations.append(max_mllt_iter)
        if not self.mllt_iterations:
            self.mllt_iterations = range(1, 4)
        self.random_prune = 4.0

        self.feature_config.lda = True
        self.feature_config.deltas = True
        self.uses_splices = True

    def compute_calculated_properties(self) -> None:
        """Generate realignment iterations, MLLT estimation iterations, and initial gaussians based on configuration"""
        super(LdaTrainer, self).compute_calculated_properties()
        self.mllt_iterations = []
        max_mllt_iter = int(self.num_iterations / 2) - 1
        for i in range(1, max_mllt_iter):
            if i < max_mllt_iter / 2 and i % 2 == 0:
                self.mllt_iterations.append(i)
        self.mllt_iterations.append(max_mllt_iter)

    @property
    def train_type(self) -> str:
        """Training identifier"""
        return "lda"

    @property
    def lda_options(self) -> MetaDict:
        """Options for computing LDA"""
        return {
            "lda_dimension": self.lda_dimension,
            "boost_silence": self.boost_silence,
            "random_prune": self.random_prune,
            "silence_csl": self.dictionary.silence_csl,
        }

    def init_training(
        self,
        identifier: str,
        temporary_directory: str,
        corpus: Corpus,
        dictionary: DictionaryType,
        previous_trainer: Optional[TrainerType],
    ):
        """
        Initialize LDA training

        Parameters
        ----------
        identifier: str
            Identifier for the training block
        temporary_directory: str
            Root temporary directory to save
        corpus: :class:`~montreal_forced_aligner.corpus.base.Corpus`
            Corpus to use
        dictionary: DictionaryType
            Dictionary to use
        previous_trainer: TrainerType, optional
            Previous trainer to initialize from

        Raises
        ------
        :class:`~montreal_forced_aligner.exceptions.KaldiProcessingError`
            If there were any errors in running Kaldi binaries
        """
        self._setup_for_init(identifier, temporary_directory, corpus, dictionary, previous_trainer)
        done_path = os.path.join(self.train_directory, "done")
        dirty_path = os.path.join(self.train_directory, "dirty")
        if os.path.exists(done_path):
            self.logger.info("{self.identifier} training already done, skipping initialization.")
            return
        begin = time.time()
        try:
            self.feature_config.directory = None
            lda_acc_stats(self)
            self.feature_config.directory = self.train_directory
        except Exception as e:
            with open(dirty_path, "w") as _:
                pass
            if isinstance(e, KaldiProcessingError):
                log_kaldi_errors(e.error_logs, self.logger)
                e.update_log_file(self.logger.handlers[0].baseFilename)
            raise
        self._setup_tree()
        self.iteration = 1
        self.logger.info("Initialization complete!")
        self.logger.debug(f"Initialization took {time.time() - begin} seconds")

    def training_iteration(self):
        """
        Run a single training iteration
        """
        if os.path.exists(self.next_model_path):
            return
        if self.iteration in self.realignment_iterations:
            align(self)
            if self.debug:
                compute_alignment_improvement(self)
        if self.iteration in self.mllt_iterations:
            calc_lda_mllt(self)

        acc_stats(self)
        parse_logs(self.log_directory)
        if self.iteration < self.final_gaussian_iteration:
            self.increment_gaussians()
        self.iteration += 1



"""Multiprocessing functions for ivector extraction and training"""
from __future__ import annotations

import os
import subprocess
from typing import TYPE_CHECKING, Dict, List, Union

from ..helper import load_scp
from ..utils import thirdparty_binary
from .helper import run_mp, run_non_mp

if TYPE_CHECKING:
    from ..config import ConfigDict
    from ..corpus.classes import File, Speaker, Utterance  # noqa
    from ..segmenter import SegmentationType, Segmenter
    from ..speaker_classifier import SpeakerClassifier
    from ..trainers.ivector_extractor import IvectorExtractorTrainer


__all__ = [
    "gmm_gselect",
    "acc_global_stats",
    "gauss_to_post",
    "acc_ivector_stats",
    "extract_ivectors",
    "get_initial_segmentation",
    "merge_segments",
    "segment_vad",
]


def gmm_gselect_func(
    log_path: str,
    dictionaries: List[str],
    feature_strings: Dict[str, str],
    ivector_options: ConfigDict,
    dubm_path: str,
    gselect_paths: Dict[str, str],
) -> None:
    """
    Multiprocessing function for running gmm-gselect

    Parameters
    ----------
    log_path: str
        Path to save log output
    dictionaries: List[str]
        List of dictionary names
    feature_strings: Dict[str, str]
        Dictionary of feature strings per dictionary name
    ivector_options: ConfigDict
        Options for ivector extractor training
    dubm_path: str
        Path to the DUBM file
    gselect_paths: Dict[str, str]
        Dictionary of gselect archives per dictionary name
    """
    with open(log_path, "w", encoding="utf8") as log_file:
        for dict_name in dictionaries:
            feature_string = feature_strings[dict_name]
            gselect_path = gselect_paths[dict_name]
            subsample_feats_proc = subprocess.Popen(
                [
                    thirdparty_binary("subsample-feats"),
                    f"--n={ivector_options['subsample']}",
                    feature_string,
                    "ark:-",
                ],
                stdout=subprocess.PIPE,
                stderr=log_file,
                env=os.environ,
            )

            gselect_proc = subprocess.Popen(
                [
                    thirdparty_binary("gmm-gselect"),
                    f"--n={ivector_options['num_gselect']}",
                    dubm_path,
                    "ark:-",
                    f"ark:{gselect_path}",
                ],
                stdin=subsample_feats_proc.stdout,
                stderr=log_file,
                env=os.environ,
            )
            gselect_proc.communicate()


def gmm_gselect(trainer: IvectorExtractorTrainer) -> None:
    """
    Multiprocessing function that stores Gaussian selection indices on disk

    See:

    - http://kaldi-asr.org/doc/gmm-gselect_8cc.html

    for more details
    on the Kaldi binary this runs.

    Also see https://github.com/kaldi-asr/kaldi/blob/master/egs/wsj/s5/steps/train_diag_ubm.sh
    for the original bash script that this function was based on.

    Parameters
    ----------
    trainer : :class:`~montreal_forced_aligner.trainers.ivector.IvectorExtractorTrainer`
        Ivector Extractor Trainer
    """
    jobs = [x.gmm_gselect_arguments(trainer) for x in trainer.corpus.jobs]
    if trainer.use_mp:
        run_mp(gmm_gselect_func, jobs, trainer.working_log_directory)
    else:
        run_non_mp(gmm_gselect_func, jobs, trainer.working_log_directory)


def acc_global_stats_func(
    log_path: str,
    dictionaries: List[str],
    feature_strings: Dict[str, str],
    ivector_options: ConfigDict,
    gselect_paths: Dict[str, str],
    acc_paths: Dict[str, str],
    dubm_path: str,
) -> None:
    """
    Multiprocessing function for accumulating global model stats

    Parameters
    ----------
    log_path: str
        Path to save log output
    dictionaries: List[str]
        List of dictionary names
    feature_strings: Dict[str, str]
        Dictionary of feature strings per dictionary name
    ivector_options: ConfigDict
        Options for ivector extractor training
    gselect_paths: Dict[str, str]
        Dictionary of gselect archives per dictionary name
    acc_paths: Dict[str, str]
        Dictionary of accumulated stats files per dictionary name
    dubm_path: str
        Path to the DUBM file
    """
    with open(log_path, "w", encoding="utf8") as log_file:
        for dict_name in dictionaries:
            feature_string = feature_strings[dict_name]
            gselect_path = gselect_paths[dict_name]
            acc_path = acc_paths[dict_name]
            subsample_feats_proc = subprocess.Popen(
                [
                    thirdparty_binary("subsample-feats"),
                    f"--n={ivector_options['subsample']}",
                    feature_string,
                    "ark:-",
                ],
                stdout=subprocess.PIPE,
                stderr=log_file,
                env=os.environ,
            )
            gmm_global_acc_proc = subprocess.Popen(
                [
                    thirdparty_binary("gmm-global-acc-stats"),
                    f"--gselect=ark:{gselect_path}",
                    dubm_path,
                    "ark:-",
                    acc_path,
                ],
                stderr=log_file,
                stdin=subsample_feats_proc.stdout,
                env=os.environ,
            )
            gmm_global_acc_proc.communicate()


def acc_global_stats(trainer: IvectorExtractorTrainer) -> None:
    """
    Multiprocessing function that accumulates global GMM stats

    See:

    - http://kaldi-asr.org/doc/gmm-global-acc-stats_8cc.html

    for more details
    on the Kaldi binary this runs.

    Also see https://github.com/kaldi-asr/kaldi/blob/master/egs/wsj/s5/steps/train_diag_ubm.sh
    for the original bash script that this function was based on.

    Parameters
    ----------
    trainer : :class:`~montreal_forced_aligner.trainers.ivector.IvectorExtractorTrainer`
        Ivector Extractor Trainer
    """
    jobs = [x.acc_global_stats_arguments(trainer) for x in trainer.corpus.jobs]
    if trainer.use_mp:
        run_mp(acc_global_stats_func, jobs, trainer.working_log_directory)
    else:
        run_non_mp(acc_global_stats_func, jobs, trainer.working_log_directory)

    # Don't remove low-count Gaussians till the last tier,
    # or gselect info won't be valid anymore
    if trainer.iteration < trainer.ubm_num_iterations:
        opt = "--remove-low-count-gaussians=false"
    else:
        opt = f"--remove-low-count-gaussians={trainer.ubm_remove_low_count_gaussians}"
    log_path = os.path.join(trainer.working_log_directory, f"update.{trainer.iteration}.log")
    with open(log_path, "w") as log_file:
        acc_files = []
        for j in jobs:
            acc_files.extend(j.acc_paths.values())
        sum_proc = subprocess.Popen(
            [thirdparty_binary("gmm-global-sum-accs"), "-"] + acc_files,
            stderr=log_file,
            stdout=subprocess.PIPE,
            env=os.environ,
        )
        gmm_global_est_proc = subprocess.Popen(
            [
                thirdparty_binary("gmm-global-est"),
                opt,
                f"--min-gaussian-weight={trainer.ubm_min_gaussian_weight}",
                trainer.current_dubm_path,
                "-",
                trainer.next_dubm_path,
            ],
            stderr=log_file,
            stdin=sum_proc.stdout,
            env=os.environ,
        )
        gmm_global_est_proc.communicate()
        # Clean up
        if not trainer.debug:
            for p in acc_files:
                os.remove(p)


def gauss_to_post_func(
    log_path: str,
    dictionaries: List[str],
    feature_strings: Dict[str, str],
    ivector_options: ConfigDict,
    post_paths: Dict[str, str],
    dubm_path: str,
):
    """
    Multiprocessing function to get posteriors during UBM training

    Parameters
    ----------
    log_path: str
        Path to save log output
    dictionaries: List[str]
        List of dictionary names
    feature_strings: Dict[str, str]
        Dictionary of feature strings per dictionary name
    ivector_options: ConfigDict
        Options for ivector extractor training
    post_paths: Dict[str, str]
        Dictionary of posterior archives per dictionary name
    dubm_path: str
        Path to the DUBM file
    """
    modified_posterior_scale = ivector_options["posterior_scale"] * ivector_options["subsample"]
    with open(log_path, "w", encoding="utf8") as log_file:
        for dict_name in dictionaries:
            feature_string = feature_strings[dict_name]
            post_path = post_paths[dict_name]
            subsample_feats_proc = subprocess.Popen(
                [
                    thirdparty_binary("subsample-feats"),
                    f"--n={ivector_options['subsample']}",
                    feature_string,
                    "ark:-",
                ],
                stdout=subprocess.PIPE,
                stderr=log_file,
                env=os.environ,
            )
            gmm_global_get_post_proc = subprocess.Popen(
                [
                    thirdparty_binary("gmm-global-get-post"),
                    f"--n={ivector_options['num_gselect']}",
                    f"--min-post={ivector_options['min_post']}",
                    dubm_path,
                    "ark:-",
                    "ark:-",
                ],
                stdout=subprocess.PIPE,
                stdin=subsample_feats_proc.stdout,
                stderr=log_file,
                env=os.environ,
            )
            scale_post_proc = subprocess.Popen(
                [
                    thirdparty_binary("scale-post"),
                    "ark:-",
                    str(modified_posterior_scale),
                    f"ark:{post_path}",
                ],
                stdin=gmm_global_get_post_proc.stdout,
                stderr=log_file,
                env=os.environ,
            )
            scale_post_proc.communicate()


def gauss_to_post(trainer: IvectorExtractorTrainer) -> None:
    """
    Multiprocessing function that does Gaussian selection and posterior extraction

    See:

    - http://kaldi-asr.org/doc/gmm-global-get-post_8cc.html
    - http://kaldi-asr.org/doc/scale-post_8cc.html

    for more details
    on the Kaldi binary this runs.

    Also see https://github.com/kaldi-asr/kaldi/blob/master/egs/wsj/s5/steps/online/nnet2/train_ivector_extractor.sh
    for the original bash script that this function was based on.

    Parameters
    ----------
    trainer: :class:`~montreal_forced_aligner.trainers.ivector.IvectorExtractorTrainer`
        Ivector Extractor Trainer
    """
    jobs = [x.gauss_to_post_arguments(trainer) for x in trainer.corpus.jobs]
    if trainer.use_mp:
        run_mp(gauss_to_post_func, jobs, trainer.working_log_directory)
    else:
        run_non_mp(gauss_to_post_func, jobs, trainer.working_log_directory)


def acc_ivector_stats_func(
    log_path: str,
    dictionaries: List[str],
    feature_strings: Dict[str, str],
    ivector_options: ConfigDict,
    ie_path: str,
    post_paths: Dict[str, str],
    acc_init_paths: Dict[str, str],
) -> None:
    """
    Multiprocessing function that accumulates stats for ivector training

    Parameters
    ----------
    log_path: str
        Path to save log output
    dictionaries: List[str]
        List of dictionary names
    feature_strings: Dict[str, str]
        Dictionary of feature strings per dictionary name
    ivector_options: ConfigDict
        Options for ivector extractor training
    ie_path: str
        Path to the ivector extractor file
    post_paths: Dict[str, str]
        Dictionary of posterior archives per dictionary name
    acc_init_paths: Dict[str, str]
        Dictionary of accumulated stats files per dictionary name
    """
    with open(log_path, "w", encoding="utf8") as log_file:
        for dict_name in dictionaries:
            feature_string = feature_strings[dict_name]
            post_path = post_paths[dict_name]
            acc_init_path = acc_init_paths[dict_name]
            subsample_feats_proc = subprocess.Popen(
                [
                    thirdparty_binary("subsample-feats"),
                    f"--n={ivector_options['subsample']}",
                    feature_string,
                    "ark:-",
                ],
                stdout=subprocess.PIPE,
                stderr=log_file,
                env=os.environ,
            )
            acc_stats_proc = subprocess.Popen(
                [
                    thirdparty_binary("ivector-extractor-acc-stats"),
                    "--num-threads=1",
                    ie_path,
                    "ark:-",
                    f"ark:{post_path}",
                    acc_init_path,
                ],
                stdin=subsample_feats_proc.stdout,
                stderr=log_file,
                env=os.environ,
            )
            acc_stats_proc.communicate()


def acc_ivector_stats(trainer: IvectorExtractorTrainer) -> None:
    """
    Multiprocessing function that calculates job_name-vector extractor stats

    See:

    - http://kaldi-asr.org/doc/ivector-extractor-acc-stats_8cc.html
    - http://kaldi-asr.org/doc/ivector-extractor-sum-accs_8cc.html

    for more details
    on the Kaldi binary this runs.

    Also see https://github.com/kaldi-asr/kaldi/blob/master/egs/wsj/s5/steps/online/nnet2/train_ivector_extractor.sh
    for the original bash script that this function was based on.

    Parameters
    ----------
    trainer: :class:`~montreal_forced_aligner.trainers.ivector.IvectorExtractorTrainer`
        Ivector Extractor Trainer
    """

    jobs = [x.ivector_acc_stats_arguments(trainer) for x in trainer.corpus.jobs]
    if trainer.use_mp:
        run_mp(acc_ivector_stats_func, jobs, trainer.working_log_directory)
    else:
        run_non_mp(acc_ivector_stats_func, jobs, trainer.working_log_directory)

    log_path = os.path.join(trainer.working_log_directory, f"sum_acc.{trainer.iteration}.log")
    acc_path = os.path.join(trainer.working_directory, f"acc.{trainer.iteration}")
    with open(log_path, "w", encoding="utf8") as log_file:
        accinits = []
        for j in jobs:
            accinits.extend(j.acc_init_paths.values())
        sum_accs_proc = subprocess.Popen(
            [thirdparty_binary("ivector-extractor-sum-accs"), "--parallel=true"]
            + accinits
            + [acc_path],
            stderr=log_file,
            env=os.environ,
        )

        sum_accs_proc.communicate()
    # clean up
    for p in accinits:
        os.remove(p)
        # Est extractor
    log_path = os.path.join(trainer.working_log_directory, f"update.{trainer.iteration}.log")
    with open(log_path, "w") as log_file:
        extractor_est_proc = subprocess.Popen(
            [
                thirdparty_binary("ivector-extractor-est"),
                f"--num-threads={trainer.corpus.num_jobs}",
                f"--gaussian-min-count={trainer.gaussian_min_count}",
                trainer.current_ie_path,
                os.path.join(trainer.working_directory, f"acc.{trainer.iteration}"),
                trainer.next_ie_path,
            ],
            stderr=log_file,
            env=os.environ,
        )
        extractor_est_proc.communicate()


def extract_ivectors_func(
    log_path: str,
    dictionaries: List[str],
    feature_strings: Dict[str, str],
    ivector_options: ConfigDict,
    ali_paths: Dict[str, str],
    ie_path: str,
    ivector_paths: Dict[str, str],
    weight_paths: Dict[str, str],
    model_path: str,
    dubm_path: str,
) -> None:
    """
    Multiprocessing function for extracting ivectors

    Parameters
    ----------
    log_path: str
        Path to save log output
    dictionaries: List[str]
        List of dictionary names
    feature_strings: Dict[str, str]
        Dictionary of feature strings per dictionary name
    ivector_options: ConfigDict
        Options for ivector extraction
    ali_paths: Dict[str, str]
        Dictionary of alignment archives per dictionary name
    ie_path: str
        Path to the ivector extractor file
    ivector_paths: Dict[str, str]
        Dictionary of ivector archives per dictionary name
    weight_paths: Dict[str, str]
        Dictionary of weighted archives per dictionary name
    model_path: str
        Path to the acoustic model file
    dubm_path: str
        Path to the DUBM file
    """
    with open(log_path, "w", encoding="utf8") as log_file:
        for dict_name in dictionaries:
            ali_path = ali_paths[dict_name]
            weight_path = weight_paths[dict_name]
            ivectors_path = ivector_paths[dict_name]
            feature_string = feature_strings[dict_name]
            use_align = os.path.exists(ali_path)
            if use_align:
                ali_to_post_proc = subprocess.Popen(
                    [thirdparty_binary("ali-to-post"), f"ark:{ali_path}", "ark:-"],
                    stderr=log_file,
                    stdout=subprocess.PIPE,
                    env=os.environ,
                )
                weight_silence_proc = subprocess.Popen(
                    [
                        thirdparty_binary("weight-silence-post"),
                        str(ivector_options["silence_weight"]),
                        ivector_options["sil_phones"],
                        model_path,
                        "ark:-",
                        "ark:-",
                    ],
                    stderr=log_file,
                    stdin=ali_to_post_proc.stdout,
                    stdout=subprocess.PIPE,
                    env=os.environ,
                )
                post_to_weight_proc = subprocess.Popen(
                    [thirdparty_binary("post-to-weights"), "ark:-", f"ark:{weight_path}"],
                    stderr=log_file,
                    stdin=weight_silence_proc.stdout,
                    env=os.environ,
                )
                post_to_weight_proc.communicate()

            gmm_global_get_post_proc = subprocess.Popen(
                [
                    thirdparty_binary("gmm-global-get-post"),
                    f"--n={ivector_options['num_gselect']}",
                    f"--min-post={ivector_options['min_post']}",
                    dubm_path,
                    feature_string,
                    "ark:-",
                ],
                stdout=subprocess.PIPE,
                stderr=log_file,
                env=os.environ,
            )
            if use_align:
                weight_proc = subprocess.Popen(
                    [
                        thirdparty_binary("weight-post"),
                        "ark:-",
                        f"ark,s,cs:{weight_path}",
                        "ark:-",
                    ],
                    stdin=gmm_global_get_post_proc.stdout,
                    stdout=subprocess.PIPE,
                    stderr=log_file,
                    env=os.environ,
                )
                extract_in = weight_proc.stdout
            else:
                extract_in = gmm_global_get_post_proc.stdout
            extract_proc = subprocess.Popen(
                [
                    thirdparty_binary("ivector-extract"),
                    f"--acoustic-weight={ivector_options['posterior_scale']}",
                    "--compute-objf-change=true",
                    f"--max-count={ivector_options['max_count']}",
                    ie_path,
                    feature_string,
                    "ark,s,cs:-",
                    f"ark,t:{ivectors_path}",
                ],
                stderr=log_file,
                stdin=extract_in,
                env=os.environ,
            )
            extract_proc.communicate()


def extract_ivectors(ivector_extractor: Union[SpeakerClassifier, IvectorExtractorTrainer]) -> None:
    """
    Multiprocessing function that extracts job_name-vectors.

    See:

    - http://kaldi-asr.org/doc/ivector-extract-online2_8cc.html
    - http://kaldi-asr.org/doc/copy-feats_8cc.html

    for more details
    on the Kaldi binary this runs.

    Also see https://github.com/kaldi-asr/kaldi/blob/master/egs/wsj/s5/steps/online/nnet2/extract_ivectors_online.sh
    for the original bash script that this function was based on.

    Parameters
    ----------
    ivector_extractor: :class:`~montreal_forced_aligner.speaker_classifier.SpeakerClassifier` or :class:`~montreal_forced_aligner.trainers.ivector.IvectorExtractorTrainer`
        Ivector extractor
    """

    log_dir = ivector_extractor.log_directory
    os.makedirs(log_dir, exist_ok=True)

    jobs = [x.extract_ivector_arguments(ivector_extractor) for x in ivector_extractor.corpus.jobs]
    if ivector_extractor.use_mp:
        run_mp(extract_ivectors_func, jobs, log_dir)
    else:
        run_non_mp(extract_ivectors_func, jobs, log_dir)


def get_initial_segmentation(frames: List[Union[int, str]], frame_shift: int) -> SegmentationType:
    """
    Compute initial segmentation over voice activity

    Parameters
    ----------
    frames: List[Union[int, str]]
        List of frames with VAD output
    frame_shift: int
        Frame shift of features in ms

    Returns
    -------
    SegmentationType
        Initial segmentation
    """
    segs = []
    cur_seg = None
    silent_frames = 0
    non_silent_frames = 0
    for i, f in enumerate(frames):
        if int(f) > 0:
            non_silent_frames += 1
            if cur_seg is None:
                cur_seg = {"begin": i * frame_shift}
        else:
            silent_frames += 1
            if cur_seg is not None:
                cur_seg["end"] = (i - 1) * frame_shift
                segs.append(cur_seg)
                cur_seg = None
    if cur_seg is not None:
        cur_seg["end"] = len(frames) * frame_shift
        segs.append(cur_seg)
    return segs


def merge_segments(
    segments: SegmentationType,
    min_pause_duration: float,
    max_segment_length: float,
    snap_boundary_threshold: float,
) -> SegmentationType:
    """
    Merge segments together

    Parameters
    ----------
    segments: SegmentationType
        Initial segments
    min_pause_duration: float
        Minimum amount of silence time to mark an utterance boundary
    max_segment_length: float
        Maximum length of segments before they're broken up
    snap_boundary_threshold:
        Boundary threshold to snap boundaries together

    Returns
    -------
    SegmentationType
        Merged segments
    """
    merged_segs = []
    for s in segments:
        if (
            not merged_segs
            or s["begin"] > merged_segs[-1]["end"] + min_pause_duration
            or s["end"] - merged_segs[-1]["begin"] > max_segment_length
        ):
            if s["end"] - s["begin"] > min_pause_duration:
                if merged_segs and snap_boundary_threshold:
                    boundary_gap = s["begin"] - merged_segs[-1]["end"]
                    if boundary_gap < snap_boundary_threshold:
                        half_boundary = boundary_gap / 2
                    else:
                        half_boundary = snap_boundary_threshold / 2
                    merged_segs[-1]["end"] += half_boundary
                    s["begin"] -= half_boundary

                merged_segs.append(s)
        else:
            merged_segs[-1]["end"] = s["end"]
    return merged_segs


def segment_vad_func(
    dictionaries: List[str],
    vad_paths: Dict[str, str],
    segmentation_options: ConfigDict,
) -> Dict[str, Utterance]:
    """
    Multiprocessing function to generate segments from VAD output

    Parameters
    ----------
    log_path: str
        Path to save log output
    dictionaries: List[str]
        List of dictionary names
    vad_paths: Dict[str, str]
        Dictionary of VAD archives per dictionary name
    segmentation_options: ConfigDict
        Options for segmentation
    """

    utterances = {}
    from ..corpus.classes import File, Speaker, Utterance  # noqa

    speaker = Speaker("speech")
    for dict_name in dictionaries:
        vad_path = vad_paths[dict_name]

        vad = load_scp(vad_path, data_type=int)
        for recording, frames in vad.items():
            file = File(recording)
            initial_segments = get_initial_segmentation(
                frames, segmentation_options["frame_shift"]
            )
            merged = merge_segments(
                initial_segments,
                segmentation_options["min_pause_duration"],
                segmentation_options["max_segment_length"],
                segmentation_options["snap_boundary_threshold"],
            )
            for seg in merged:
                utterances[recording] = Utterance(
                    speaker, file, begin=seg["begin"], end=seg["end"], text="speech"
                )
    return utterances


def segment_vad(segmenter: Segmenter) -> None:
    """
    Run segmentation based off of VAD

    Parameters
    ----------
    segmenter: :class:`~montreal_forced_aligner.segmenter.Segmenter`
        Segmenter
    """

    from ..corpus.classes import Speaker  # noqa

    jobs = [x.segments_vad_arguments(segmenter) for x in segmenter.corpus.jobs]
    if segmenter.segmentation_config.use_mp:
        segment_info = run_mp(
            segment_vad_func, jobs, segmenter.corpus.features_log_directory, True
        )
    else:
        segment_info = run_non_mp(
            segment_vad_func, jobs, segmenter.corpus.features_log_directory, True
        )
    for j in segmenter.corpus.jobs:
        for old_utt, utterance in segment_info[j.name].items():
            old_utt = segmenter.corpus.utterances[old_utt]
            file = old_utt.file
            if segmenter.corpus.no_speakers:
                if utterance.speaker_name not in segmenter.corpus.speakers:
                    segmenter.corpus.speakers[utterance.speaker_name] = Speaker(
                        utterance.speaker_name
                    )
                speaker = segmenter.corpus.speakers[utterance.speaker_name]
            else:
                speaker = old_utt.speaker
            utterance.file = file
            utterance.set_speaker(speaker)
            segmenter.corpus.add_utterance(utterance)
    utterance_ids = [x.name for x in segmenter.corpus.utterances.values() if x.begin is None]
    for u in utterance_ids:
        segmenter.corpus.delete_utterance(u)



"""Class definitions for configuring acoustic model training"""
from __future__ import annotations

import os
from collections import Counter
from typing import Iterator, List, Tuple

import yaml

from ..trainers import (
    BaseTrainer,
    IvectorExtractorTrainer,
    LdaTrainer,
    MonophoneTrainer,
    SatTrainer,
    TriphoneTrainer,
)
from .align_config import AlignConfig
from .base_config import (
    DEFAULT_CLITIC_MARKERS,
    DEFAULT_COMPOUND_MARKERS,
    DEFAULT_PUNCTUATION,
    PARSING_KEYS,
    BaseConfig,
    ConfigError,
)
from .feature_config import FeatureConfig

__all__ = [
    "TrainingConfig",
    "train_yaml_to_config",
    "load_basic_train",
    "load_basic_train_ivector",
    "load_test_config",
    "load_sat_adapt",
    "load_no_sat_adapt",
]


class TrainingConfig(BaseConfig):
    """
    Configuration class for storing parameters and trainers for training acoustic models
    """

    def __init__(self, training_configs):
        self.training_configs = training_configs
        counts = Counter([x.train_type for x in self.training_configs])
        self.training_identifiers = []
        curs = {x.train_type: 1 for x in self.training_configs}
        for t in training_configs:
            i = t.train_type
            if counts[t.train_type] != 1:
                i += str(curs[t.train_type])
                curs[t.train_type] += 1
            self.training_identifiers.append(i)

        self.punctuation = DEFAULT_PUNCTUATION
        self.clitic_markers = DEFAULT_CLITIC_MARKERS
        self.compound_markers = DEFAULT_COMPOUND_MARKERS

    def update_from_align(self, align_config: AlignConfig) -> None:
        """Update parameters from an AlignConfig"""
        for tc in self.training_configs:
            tc.overwrite = align_config.overwrite
            tc.cleanup_textgrids = align_config.cleanup_textgrids

    def update(self, data: dict) -> None:
        """Update parameters"""
        for k, v in data.items():
            if k in PARSING_KEYS:
                if not v:
                    continue
                if "-" in v:
                    v = "-" + v.replace("-", "")
                if "]" in v and r"\]" not in v:
                    v = v.replace("]", r"\]")
            if not hasattr(self, k):
                raise ConfigError("No field found for key {}".format(k))
            setattr(self, k, v)
        for trainer in self.values():
            trainer.update(data)

    def keys(self) -> List:
        """List of training identifiers"""
        return self.training_identifiers

    def values(self) -> List[BaseTrainer]:
        """List of trainers"""
        return self.training_configs

    def items(self) -> Iterator:
        """Iterator over training identifiers and trainers"""
        return zip(self.training_identifiers, self.training_configs)

    def __getitem__(self, item: str) -> BaseTrainer:
        """Get trainer based on identifier"""
        if item not in self.training_identifiers:
            raise KeyError(f"{item} not a valid training identifier")
        return self.training_configs[self.training_identifiers.index(item)]

    @property
    def uses_sat(self) -> bool:
        """Flag for whether a trainer uses speaker adaptation"""
        for k in self.keys():
            if k.startswith("sat"):
                return True
        return False


def train_yaml_to_config(
    path: str, require_mono: bool = True
) -> Tuple[TrainingConfig, AlignConfig]:
    """
    Helper function to load acoustic model training configurations

    Parameters
    ----------
    path: str
        Path to yaml file

    Returns
    -------
    :class:`~montreal_forced_aligner.config.train_config.TrainingConfig`
        Training configuration
    :class:`~montreal_forced_aligner.config.align_config.AlignConfig`
        Alignment configuration
    """
    with open(path, "r", encoding="utf8") as f:
        data = yaml.load(f, Loader=yaml.SafeLoader)
        global_params = {}
        training = []
        training_params = []
        global_feature_params = {}
        for k, v in data.items():
            if k == "training":
                for t in v:
                    for k2, v2 in t.items():
                        feature_config = FeatureConfig()
                        if k2 == "monophone":
                            training.append(MonophoneTrainer(feature_config))
                        elif k2 == "triphone":
                            training.append(TriphoneTrainer(feature_config))
                        elif k2 == "lda":
                            training.append(LdaTrainer(feature_config))
                        elif k2 == "sat":
                            training.append(SatTrainer(feature_config))
                        elif k2 == "ivector":
                            training.append(IvectorExtractorTrainer(feature_config))
                        training_params.append(v2)
            elif k == "features":
                global_feature_params.update(v)
            else:
                global_params[k] = v
        feature_config = FeatureConfig()
        feature_config.update(global_feature_params)
        align_config = AlignConfig(feature_config)
        align_config.update(global_params)
        training_config = None
        if training:
            for i, t in enumerate(training):
                if i == 0 and require_mono and t.train_type not in ["mono", "ivector"]:
                    raise ConfigError("The first round of training must be monophone.")
                t.update(global_params)
                t.update(training_params[i])
                t.feature_config.update(global_feature_params)
            training_config = TrainingConfig(training)
        align_config.feature_config.fmllr = training_config.uses_sat
        if align_config.beam >= align_config.retry_beam:
            raise ConfigError("Retry beam must be greater than beam.")
        return training_config, align_config


def load_basic_train() -> Tuple[TrainingConfig, AlignConfig]:
    """
    Helper function to load the default parameters

    Returns
    -------
    :class:`~montreal_forced_aligner.config.train_config.TrainingConfig`
        Training configuration
    :class:`~montreal_forced_aligner.config.align_config.AlignConfig`
        Alignment configuration
    """
    base_dir = os.path.dirname(os.path.abspath(__file__))
    training_config, align_config = train_yaml_to_config(
        os.path.join(base_dir, "basic_train.yaml")
    )
    return training_config, align_config


def load_sat_adapt() -> Tuple[TrainingConfig, AlignConfig]:
    """
    Helper function to load the default speaker adaptation parameters for adapting an acoustic model to new data

    Returns
    -------
    :class:`~montreal_forced_aligner.config.train_config.TrainingConfig`
        Training configuration
    :class:`~montreal_forced_aligner.config.align_config.AlignConfig`
        Alignment configuration
    """
    base_dir = os.path.dirname(os.path.abspath(__file__))
    training_config, align_config = train_yaml_to_config(
        os.path.join(base_dir, "adapt_sat.yaml"), require_mono=False
    )
    training_config.training_configs[0].fmllr_iterations = range(
        0, training_config.training_configs[0].num_iterations
    )
    training_config.training_configs[0].realignment_iterations = range(
        0, training_config.training_configs[0].num_iterations
    )
    return training_config, align_config


def load_no_sat_adapt() -> Tuple[TrainingConfig, AlignConfig]:
    """
    Helper function to load the default parameters for adapting an acoustic model to new data without speaker adaptation

    Returns
    -------
    :class:`~montreal_forced_aligner.config.train_config.TrainingConfig`
        Training configuration
    :class:`~montreal_forced_aligner.config.align_config.AlignConfig`
        Alignment configuration
    """
    base_dir = os.path.dirname(os.path.abspath(__file__))
    training_config, align_config = train_yaml_to_config(
        os.path.join(base_dir, "adapt_nosat.yaml"), require_mono=False
    )
    training_config.training_configs[0].realignment_iterations = range(
        0, training_config.training_configs[0].num_iterations
    )
    return training_config, align_config


def load_basic_train_ivector() -> Tuple[TrainingConfig, AlignConfig]:
    """
    Helper function to load the default parameters for training ivector extractors

    Returns
    -------
    :class:`~montreal_forced_aligner.config.train_config.TrainingConfig`
        Training configuration
    :class:`~montreal_forced_aligner.config.align_config.AlignConfig`
        Alignment configuration
    """
    base_dir = os.path.dirname(os.path.abspath(__file__))
    training_config, align_config = train_yaml_to_config(
        os.path.join(base_dir, "basic_train_ivector.yaml")
    )
    return training_config, align_config


def load_test_config() -> Tuple[TrainingConfig, AlignConfig]:
    """
    Helper function to load the default parameters for validating corpora

    Returns
    -------
    :class:`~montreal_forced_aligner.config.train_config.TrainingConfig`
        Training configuration
    :class:`~montreal_forced_aligner.config.align_config.AlignConfig`
        Alignment configuration
    """
    base_dir = os.path.dirname(os.path.abspath(__file__))
    training_config, align_config = train_yaml_to_config(
        os.path.join(base_dir, "test_config.yaml")
    )
    return training_config, align_config



"""Helper functions for MFA"""
from __future__ import annotations

import sys
import textwrap
from typing import TYPE_CHECKING, Any, Collection, List, Optional, Tuple, Type

import numpy
from colorama import Fore, Style

if TYPE_CHECKING:
    from .config import ConfigDict
    from .corpus import CorpusMappingType, ScpType

Labels = List[Any]

__all__ = [
    "TerminalPrinter",
    "comma_join",
    "make_safe",
    "make_scp_safe",
    "save_scp",
    "load_scp",
    "load_scp_safe",
    "score",
    "edit_distance",
    "output_mapping",
]


class TerminalPrinter:
    """
    Helper class to output colorized text

    Attributes
    ----------
    colors: Dict[str,str]
        Mapping of color names to terminal codes in colorama (or empty strings
        if the global terminal_colors flag is set to False)
    """

    def __init__(self):
        from .config import load_global_config

        c = load_global_config()
        self.colors = {}
        self.colors["bright"] = ""
        self.colors["green"] = ""
        self.colors["red"] = ""
        self.colors["blue"] = ""
        self.colors["cyan"] = ""
        self.colors["yellow"] = ""
        self.colors["reset"] = ""
        self.colors["normal"] = ""
        self.width = c["terminal_width"]
        if c["terminal_colors"]:
            self.colors["bright"] = Style.BRIGHT
            self.colors["green"] = Fore.GREEN
            self.colors["red"] = Fore.RED
            self.colors["blue"] = Fore.BLUE
            self.colors["cyan"] = Fore.CYAN
            self.colors["yellow"] = Fore.YELLOW
            self.colors["reset"] = Style.RESET_ALL
            self.colors["normal"] = Style.NORMAL

    def colorize(self, text: str, color: str) -> str:
        """
        Colorize a string

        Parameters
        ----------
        text: str
            Text to colorize
        color: str
            Colorama code or empty string to wrap the text

        Returns
        -------
        str
            Colorized string
        """
        return f"{self.colors[color]}{text}{self.colors['reset']}"

    def print_block(self, block: dict, starting_level: int = 1) -> None:
        """
        Print a configuration block

        Parameters
        ----------
        block: dict
            Configuration options to output
        starting_level: int
            Starting indentation level
        """
        for k, v in block.items():
            value_color = None
            key_color = None
            value = ""
            if isinstance(k, tuple):
                k, key_color = k

            if isinstance(v, tuple):
                value, value_color = v
            elif not isinstance(v, dict):
                value = v
            self.print_information_line(k, value, key_color, value_color, starting_level)
            if isinstance(v, dict):
                self.print_block(v, starting_level=starting_level + 1)
        print()

    def print_config(self, configuration: ConfigDict) -> None:
        """
        Pretty print a configuration

        Parameters
        ----------
        configuration: ConfigDict
            Configuration to print
        """
        for k, v in configuration.items():
            if "name" in v:
                name = v["name"]
                name_color = None
                if isinstance(name, tuple):
                    name, name_color = name
                self.print_information_line(k, name, value_color=name_color, level=0)
            if "data" in v:
                self.print_block(v["data"])

    def print_information_line(
        self,
        key: str,
        value: Any,
        key_color: Optional[str] = None,
        value_color: Optional[str] = None,
        level: int = 1,
    ) -> None:
        """
        Pretty print a given configuration line

        Parameters
        ----------
        key: str
            Configuration key
        value: Any
            Configuration value
        key_color: str
            Key color
        value_color: str
            Value color
        level: int
            Indentation level
        """
        if key_color is None:
            key_color = "bright"
        if value_color is None:
            value_color = "yellow"
            if isinstance(value, bool):
                if value:
                    value_color = "green"
                else:
                    value_color = "red"
        if isinstance(value, (list, tuple, set)):
            value = comma_join([self.colorize(x, value_color) for x in sorted(value)])
        else:
            value = self.colorize(value, value_color)
        indent = ("  " * level) + "-"
        subsequent_indent = "  " * (level + 1)
        if key:
            key = f" {key}:"
            subsequent_indent += " " * (len(key))
        wrapper = textwrap.TextWrapper(
            initial_indent=indent, subsequent_indent=subsequent_indent, width=self.width
        )
        print(wrapper.fill(f"{self.colorize(key, key_color)} {value}"))


def comma_join(sequence: Collection[Any]) -> str:
    """
    Helper function to combine a list into a human-readable expression with commas and a final "and" separator

    Parameters
    ----------
    sequence: Collection[Any]
        Items to join together into a list

    Returns
    -------
    str
        Joined items in list format
    """
    if len(sequence) < 3:
        return " and ".join(sequence)
    return f"{', '.join(sequence[:-1])}, and {sequence[-1]}"


def make_safe(element: Any) -> str:
    """
    Helper function to make an element a string

    Parameters
    ----------
    element: Any
        Element to recursively turn into a string

    Returns
    -------
    str
        All elements combined into a string
    """
    if isinstance(element, list):
        return " ".join(map(make_safe, element))
    return str(element)


def make_scp_safe(string: str) -> str:
    """
    Helper function to make a string safe for saving in Kaldi scp files.  They use space as a delimiter, so
    any spaces in the string will be converted to "_MFASPACE_" to preserve them

    Parameters
    ----------
    string: str
        Text to escape

    Returns
    -------
    str
        Escaped text
    """
    return string.replace(" ", "_MFASPACE_")


def load_scp_safe(string: str) -> str:
    """
    Helper function to load previously made safe text.  All instances of "_MFASPACE_" will be converted to a
    regular space character

    Parameters
    ----------
    string: str
        String to convert

    Returns
    -------
    str
        Converted string
    """
    return string.replace("_MFASPACE_", " ")


def output_mapping(mapping: CorpusMappingType, path: str, skip_safe: bool = False) -> None:
    """
    Helper function to save mapping information (i.e., utt2spk) in Kaldi scp format

    Parameters
    ----------
    mapping: CorpusMappingType
        Mapping to output
    path: str
        Path to save mapping
    skip_safe: bool, optional
        Flag for whether to skip over making a string safe
    """
    if not mapping:
        return
    with open(path, "w", encoding="utf8") as f:
        for k in sorted(mapping.keys()):
            v = mapping[k]
            if isinstance(v, (list, set, tuple)):
                v = " ".join(map(str, v))
            elif not skip_safe:
                v = make_scp_safe(v)
            f.write(f"{make_scp_safe(k)} {v}\n")


def save_scp(
    scp: ScpType, path: str, sort: Optional[bool] = True, multiline: Optional[bool] = False
) -> None:
    """
    Helper function to save an arbitrary SCP

    Parameters
    ----------
    scp: ScpType
        SCP to save
    path: str
        File path
    sort: bool, optional
        Flag for whether the output file should be sorted
    multiline: bool, optional
        Flag for whether the SCP contains multiline data (i.e., utterance FSTs)
    """
    if sys.platform == "win32":
        newline = ""
    else:
        newline = None
    if not scp:
        return
    with open(path, "w", encoding="utf8", newline=newline) as f:
        if sort:
            scp = sorted(scp)
        for line in scp:
            if multiline:
                f.write(f"{make_safe(line[0])}\n{make_safe(line[1])}\n")
            else:
                f.write(f"{' '.join(map(make_safe, line))}\n")


def load_scp(path: str, data_type: Optional[Type] = str) -> CorpusMappingType:
    """
    Load a Kaldi script file (.scp)

    See http://kaldi-asr.org/doc/io.html#io_sec_scp_details for more information

    Parameters
    ----------
    path : str
        Path to Kaldi script file
    data_type : type
        Type to coerce the data to

    Returns
    -------
    dict
        Dictionary where the keys are the first couple and the values are all
        other columns in the script file

    """
    scp = {}
    with open(path, "r", encoding="utf8") as f:
        for line in f:
            line = line.strip()
            if line == "":
                continue
            line_list = line.split()
            key = load_scp_safe(line_list.pop(0))
            if len(line_list) == 1:
                value = data_type(line_list[0])
                if isinstance(value, str):
                    value = load_scp_safe(value)
            else:
                value = [data_type(x) for x in line_list if x not in ["[", "]"]]
            scp[key] = value
    return scp


def edit_distance(x: Labels, y: Labels) -> int:
    """
    Compute edit distance between two sets of labels

    For a more expressive version of the same, see:

         https://gist.github.com/kylebgorman/8034009

    Parameters
    ----------
    x: Labels
        First sequence to compare
    y: Lables
        Second sequence to compare

    Returns
    -------
    int
        Edit distance
    """
    idim = len(x) + 1
    jdim = len(y) + 1
    table = numpy.zeros((idim, jdim), dtype=numpy.uint8)
    table[1:, 0] = 1
    table[0, 1:] = 1
    for i in range(1, idim):
        for j in range(1, jdim):
            if x[i - 1] == y[j - 1]:
                table[i][j] = table[i - 1][j - 1]
            else:
                c1 = table[i - 1][j]
                c2 = table[i][j - 1]
                c3 = table[i - 1][j - 1]
                table[i][j] = min(c1, c2, c3) + 1
    return int(table[-1][-1])


def score(gold: Labels, hypo: (Labels, List)) -> Tuple[int, int]:
    """
    Computes sufficient statistics for LER calculation.

    Parameters
    ----------
    gold: Labels
        The reference labels
    hypo: Labels
        The hypothesized labels

    Returns
    -------
    int
        Edit distance
    int
        Length of the gold labels
    """
    if isinstance(hypo, list):
        edits = 100000
        for h in hypo:
            e = edit_distance(gold, h)
            if e < edits:
                edits = e
            if not edits:
                break
    else:
        edits = edit_distance(gold, hypo)
    return edits, len(gold)



"""Class definitions for corpora"""
from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, List, Tuple, Union

from .base import Corpus  # noqa

if TYPE_CHECKING:

    SegmentsType = Dict[str, Dict[str, Union[str, float, int]]]
    OneToOneMappingType = Dict[str, str]
    OneToManyMappingType = Dict[str, List[str]]

    CorpusMappingType = Union[OneToOneMappingType, OneToManyMappingType]
    ScpType = Union[List[Tuple[str, str]], List[Tuple[str, List[Any]]]]
    CorpusGroupedOneToOne = List[List[Tuple[str, str]]]
    CorpusGroupedOneToMany = List[List[Tuple[str, List[Any]]]]
    CorpusGroupedType = Union[CorpusGroupedOneToMany, CorpusGroupedOneToOne]

__all__ = ["Corpus"]



"""Helper classes and functions for multiprocessing"""
from __future__ import annotations

import multiprocessing as mp
import os
import sys
import traceback
from queue import Empty
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

from ..utils import parse_logs

__all__ = ["Counter", "Stopped", "ProcessWorker", "run_mp", "run_non_mp"]


class Counter(object):
    """
    Multiprocessing counter object for keeping track of progress

    Attributes
    ----------
    val: multiprocessing.Value
        Integer to increment
    lock: multiprocessing.Lock
        Lock for process safety
    """

    def __init__(self, init_val: int = 0):
        self.val = mp.Value("i", init_val)
        self.lock = mp.Lock()

    def increment(self) -> None:
        """Increment the counter"""
        with self.lock:
            self.val.value += 1

    def value(self) -> int:
        """Get the current value of the counter"""
        with self.lock:
            return self.val.value


class Stopped(object):
    """
    Multiprocessing class for detecting whether processes should stop processing and exit ASAP

    Attributes
    ----------
    val: multiprocessing.Value
        0 if not stopped, 1 if stopped
    lock: multiprocessing.Lock
        Lock for process safety
    _source: multiprocessing.Value
        1 if it was a Ctrl+C event that stopped it, 0 otherwise
    """

    def __init__(self, initval: Union[bool, int] = False):
        self.val = mp.Value("i", initval)
        self.lock = mp.Lock()
        self._source = mp.Value("i", 0)

    def stop(self) -> None:
        """Signal that work should stop asap"""
        with self.lock:
            self.val.value = True

    def stop_check(self) -> int:
        """Check whether a process should stop"""
        with self.lock:
            return self.val.value

    def set_sigint_source(self) -> None:
        """Set the source as a ctrl+c"""
        with self.lock:
            self._source.value = True

    def source(self) -> int:
        """Get the source value"""
        with self.lock:
            return self._source.value


class ProcessWorker(mp.Process):
    """
    Multiprocessing function work

    Parameters
    ----------
    job_name: int
        Integer number of job
    job_q: :class:`~multiprocessing.Queue`
        Job queue to pull arguments from
    function: Callable
        Multiprocessing function to call on arguments from job_q
    return_dict: Dict
        Dictionary for collecting errors
    stopped: :class:`~montreal_forced_aligner.multiprocess.helper.Stopped`
        Stop check
    return_info: Dict[int, Any], optional
        Optional dictionary to fill if the function should return information to main thread
    """

    def __init__(
        self,
        job_name: int,
        job_q: mp.Queue,
        function: Callable,
        return_dict: Dict,
        stopped: Stopped,
        return_info: Optional[Dict[int, Any]] = None,
    ):
        mp.Process.__init__(self)
        self.job_name = job_name
        self.function = function
        self.job_q = job_q
        self.return_dict = return_dict
        self.return_info = return_info
        self.stopped = stopped

    def run(self) -> None:
        """
        Run through the arguments in the queue apply the function to them
        """
        try:
            arguments = self.job_q.get(timeout=1)
        except Empty:
            return
        self.job_q.task_done()
        try:
            result = self.function(*arguments)
            if self.return_info is not None:
                self.return_info[self.job_name] = result
        except Exception:
            self.stopped.stop()
            self.return_dict["error"] = arguments, Exception(
                traceback.format_exception(*sys.exc_info())
            )


def run_non_mp(
    function: Callable,
    argument_list: List[Tuple[Any, ...]],
    log_directory: str,
    return_info: bool = False,
) -> Optional[Dict[Any, Any]]:
    """
    Similar to run_mp, but no additional processes are used and the jobs are evaluated in sequential order

    Parameters
    ----------
    function: Callable
        Multiprocessing function to evaluate
    argument_list: List
        List of arguments to process
    log_directory: str
        Directory that all log information from the processes goes to
    return_info: Dict, optional
        If the function returns information, supply the return dict to populate

    Returns
    -------
    Dict, optional
        If the function returns information, returns the dictionary it was supplied with
    """
    if return_info:
        info = {}
        for i, args in enumerate(argument_list):
            info[i] = function(*args)
        parse_logs(log_directory)
        return info

    for args in argument_list:
        function(*args)
    parse_logs(log_directory)


def run_mp(
    function: Callable,
    argument_list: List[Tuple[Any, ...]],
    log_directory: str,
    return_info: bool = False,
) -> Optional[Dict[int, Any]]:
    """
    Apply a function for each job in parallel

    Parameters
    ----------
    function: Callable
        Multiprocessing function to apply
    argument_list: List
        List of arguments for each job
    log_directory: str
        Directory that all log information from the processes goes to
    return_info: Dict, optional
        If the function returns information, supply the return dict to populate
    """
    from ..config import BLAS_THREADS

    os.environ["OPENBLAS_NUM_THREADS"] = f"{BLAS_THREADS}"
    os.environ["MKL_NUM_THREADS"] = f"{BLAS_THREADS}"
    stopped = Stopped()
    manager = mp.Manager()
    job_queue = manager.Queue()
    return_dict = manager.dict()
    info = None
    if return_info:
        info = manager.dict()
    for a in argument_list:
        job_queue.put(a)
    procs = []
    for i in range(len(argument_list)):
        p = ProcessWorker(i, job_queue, function, return_dict, stopped, info)
        procs.append(p)
        p.start()

    for p in procs:
        p.join()
    if "error" in return_dict:
        _, exc = return_dict["error"]
        raise exc

    parse_logs(log_directory)
    if return_info:
        return info
